@@ -1,23 +1,24 @@
 """Base Classes for custom pipelines."""
-from typing import Dict, Generic, Tuple, TypeVar, Union
+from typing import Any, ClassVar, Dict, Generic, Tuple, TypeVar, Union
 
 from typing_extensions import Self
 
+from tpcp import NOTHING
 from tpcp._algorithm import Algorithm
-from tpcp._algorithm_utils import make_action_safe, make_optimize_safe
+from tpcp._algorithm_utils import ACTION_METHOD_INDICATOR, _check_safe_run, make_action_safe
 from tpcp._dataset import DatasetT
 
 PipelineT = TypeVar("PipelineT", bound="Pipeline")
 OptimizablePipelineT = TypeVar("OptimizablePipelineT", bound="OptimizablePipeline")
 
 
-class Pipeline(Algorithm, Generic[DatasetT], _skip_validation=True):
+class Pipeline(Algorithm, Generic[DatasetT]):
     """Baseclass for all custom pipelines.
 
     To create your own custom pipeline, subclass this class and implement `run`.
     """
 
-    _action_methods: Tuple[str, ...] = ("safe_run", "run")
+    _action_methods: ClassVar[Tuple[str, str]] = ("safe_run", "run")
 
     datapoint: DatasetT
 
@@ -43,7 +44,6 @@ class Pipeline(Algorithm, Generic[DatasetT], _skip_validation=True):
         """
         raise NotImplementedError()  # pragma: no cover
 
-    @make_action_safe
     def safe_run(self, datapoint: DatasetT) -> Self:
         """Run the pipeline with some additional checks.
 
@@ -69,7 +69,10 @@ class Pipeline(Algorithm, Generic[DatasetT], _skip_validation=True):
             The class instance with all result attributes populated
 
         """
-        return self.run(datapoint)
+        run_method = self.run
+        if getattr(run_method, ACTION_METHOD_INDICATOR, False) is True:
+            return run_method(datapoint)
+        return _check_safe_run(self, run_method, datapoint)
 
     def score(self, datapoint: DatasetT) -> Union[float, Dict[str, float]]:
         """Calculate performance of the pipeline on a datapoint with reference information.
@@ -96,7 +99,7 @@ class Pipeline(Algorithm, Generic[DatasetT], _skip_validation=True):
         raise NotImplementedError()  # pragma: no cover
 
 
-class OptimizablePipeline(Pipeline[DatasetT], _skip_validation=True):
+class OptimizablePipeline(Pipeline[DatasetT]):
     """Pipeline with custom ways to optimize and/or train input parameters.
 
     OptimizablePipelines are expected to implement a concrete way to train internal models or optimize parameters.
@@ -121,7 +124,6 @@ class OptimizablePipeline(Pipeline[DatasetT], _skip_validation=True):
     optimized pipeline.
     """
 
-    @make_optimize_safe
     def self_optimize(self, dataset: DatasetT, **kwargs) -> Self:
         """Optimize the input parameters of the pipeline or algorithm using any logic.
 
@@ -131,6 +133,7 @@ class OptimizablePipeline(Pipeline[DatasetT], _skip_validation=True):
         .. note::
             The optimizations must only modify the input parameters (aka `self.clone` should retain the optimization
             results).
+            If you need to return further information, implement `self_optimize_with_info` instead.
 
         Parameters
         ----------
@@ -147,4 +150,42 @@ class OptimizablePipeline(Pipeline[DatasetT], _skip_validation=True):
             The class instance with optimized input parameters.
 
         """
-        raise NotImplementedError()  # pragma: no cover
+        try:
+            # This seems hacky, but is used to avoid infinite recursion
+            setattr(type(self), "__optimize_not_implemented__", True)
+            out = self.self_optimize_with_info(dataset, **kwargs)[0]
+            delattr(type(self), "__optimize_not_implemented__")
+            return out
+        except NotImplementedError as e:
+            raise NotImplementedError() from e
+
+    def self_optimize_with_info(self, dataset: DatasetT, **kwargs) -> Tuple[Self, Any]:
+        """Optimize the input parameters of the pipeline or algorithm using any logic.
+
+        This is equivalent to `self_optimize`, but allows you to return additional information as a second return value.
+        If you implement this method, there is no need to implement `self_optimize` as well.
+
+        Parameters
+        ----------
+        dataset
+            An instance of a :class:`tpcp.dataset.Dataset` containing one or multiple data points that can
+            be used for training.
+            The structure of the data and the available reference information will depend on the dataset.
+        kwargs
+            Additional parameters required for the optimization process.
+
+        Returns
+        -------
+        self
+            The class instance with optimized input parameters.
+        info
+            An arbitrary piece of information
+
+        """
+        try:
+            if getattr(type(self), "__optimize_not_implemented__", False):
+                delattr(type(self), "__optimize_not_implemented__")
+                raise NotImplementedError()
+            return self.self_optimize(dataset, **kwargs), NOTHING
+        except NotImplementedError as e:
+            raise NotImplementedError() from e
