@@ -6,7 +6,7 @@ from contextlib import nullcontext
 from functools import partial
 from itertools import product
 from tempfile import TemporaryDirectory
-from typing import Any, ContextManager, Dict, Generic, Iterator, List, Optional, Union
+from typing import Any, ContextManager, Dict, Generic, Iterator, List, Optional, Tuple, TypeVar, Union
 
 import numpy as np
 from joblib import Memory, delayed
@@ -17,8 +17,8 @@ from tqdm.auto import tqdm
 from typing_extensions import Self
 
 from tpcp import OptimizablePipeline
-from tpcp._algorithm_utils import OPTIMIZE_METHOD_INDICATOR, _check_safe_optimize
-from tpcp._base import _get_annotated_fields_of_type
+from tpcp._algorithm_utils import OPTIMIZE_METHOD_INDICATOR, _check_safe_optimize, _split_returns
+from tpcp._base import NOTHING, _get_annotated_fields_of_type
 from tpcp._dataset import DatasetT
 from tpcp._optimize import BaseOptimize
 from tpcp._parameters import Parameter, _ParaTypes
@@ -30,12 +30,14 @@ from tpcp._utils._general import (
     _split_hyper_and_pure_parameters,
 )
 from tpcp._utils._multiprocess import TqdmParallel
-from tpcp._utils._score import _ERROR_SCORE_TYPE, _optimize_and_score, _score
+from tpcp._utils._score import _optimize_and_score, _score
 from tpcp.exceptions import PotentialUserErrorWarning
-from tpcp.validate._scorer import ScorerTypes, ScoreTypeT, _validate_scorer
+from tpcp.validate._scorer import ScorerTypes, _validate_scorer
+
+T = TypeVar("T")
 
 
-class DummyOptimize(BaseOptimize[PipelineT, DatasetT], _skip_validation=True):
+class DummyOptimize(BaseOptimize[PipelineT, DatasetT]):
     """Provide API compatibility for SimplePipelines in optimize wrappers.
 
     This is a simple dummy Optimizer that will **not** optimize anything, but just provide the correct API so that
@@ -120,6 +122,10 @@ class Optimize(BaseOptimize[OptimizablePipelineT, DatasetT]):
         If True, we add additional checks to make sure the `self_optimize` method of the pipeline is correctly
         implemented.
         See :func:`~tpcp.make_optimize_safe` for more info.
+    optimize_with_info
+        If True, Optimize will try to call `self_optimize_with_info` by default and will fall back to `self_optimize`.
+        If you want to force optimize to use `self_optimize`, even if an implementation of `self_optimize_with_info`
+        exists, set this parameter to False.
 
     Other Parameters
     ----------------
@@ -131,19 +137,25 @@ class Optimize(BaseOptimize[OptimizablePipelineT, DatasetT]):
     optimized_pipeline_
         The optimized version of the pipeline.
         That is a copy of the input pipeline with modified params.
+    optimization_info_
+        If the optimized pipeline implements a `self_optimize_with_info` method, this parameter contains the
+        additional information provided as second return value from this method.
 
     """
 
     pipeline: Parameter[OptimizablePipelineT]
     safe_optimize: bool
+    optimize_with_info: bool
 
     optimized_pipeline_: OptimizablePipelineT
+    optimization_info_: Any
 
     def __init__(  # noqa: super-init-not-called
-        self, pipeline: OptimizablePipelineT, *, safe_optimize: bool = True
+        self, pipeline: OptimizablePipelineT, *, safe_optimize: bool = True, optimize_with_info: bool = True
     ) -> None:
         self.pipeline = pipeline
         self.safe_optimize = safe_optimize
+        self.optimize_with_info = optimize_with_info
 
     def optimize(self, dataset: DatasetT, **optimize_params: Any) -> Self:
         """Run the self-optimization defined by the pipeline.
@@ -173,20 +185,25 @@ class Optimize(BaseOptimize[OptimizablePipelineT, DatasetT]):
             )
         # We clone just to make sure runs are independent
         pipeline: OptimizablePipeline = self.pipeline.clone()
+        method = pipeline.self_optimize_with_info if self.optimize_with_info else pipeline.self_optimize
         if self.safe_optimize is True:
             # We check here, if the pipeline already has the safe decorator and if yes just call it.
-            if getattr(pipeline.self_optimize, OPTIMIZE_METHOD_INDICATOR, False) is True:
-                optimized_pipeline = pipeline.self_optimize(dataset, **optimize_params)
+            if getattr(method, OPTIMIZE_METHOD_INDICATOR, False) is True:
+                optimized_pipeline, other_info = _split_returns(method(dataset, **optimize_params))
             else:
-                optimized_pipeline = _check_safe_optimize(pipeline, pipeline.self_optimize, dataset, **optimize_params)
+                optimized_pipeline, other_info = _split_returns(
+                    _check_safe_optimize(pipeline, method, dataset, **optimize_params)
+                )
         else:
-            optimized_pipeline = pipeline.self_optimize(dataset, **optimize_params)
+            optimized_pipeline, other_info = _split_returns(method(dataset, **optimize_params))
         # We clone again, just to be sure
         self.optimized_pipeline_ = optimized_pipeline.clone()
+        if self.optimize_with_info and other_info is not NOTHING:
+            self.optimization_info_ = other_info
         return self
 
 
-class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT, ScoreTypeT]):
+class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT, T]):
     """Perform a grid search over various parameters.
 
     This scores the pipeline for every combination of data points in the provided dataset and parameter combinations
@@ -230,10 +247,6 @@ class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT,
         If False, the respective result attributes will not be populated.
         If multiple parameter combinations have the same score, the one tested first will be used.
         Otherwise, higher values are always considered better.
-    error_score
-        Value to assign to the score if an error occurs during scoring.
-        If set to ‘raise’, the error is raised.
-        If a numeric value is given, a Warning is raised.
     progress_bar
         True/False to enable/disable a tqdm progress bar.
 
@@ -286,11 +299,10 @@ class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT,
     """
 
     parameter_grid: ParameterGrid
-    scoring: ScorerTypes[PipelineT, DatasetT, ScoreTypeT]
+    scoring: ScorerTypes[PipelineT, DatasetT, T]
     n_jobs: Optional[int]
     return_optimized: Union[bool, str]
     pre_dispatch: Union[int, str]
-    error_score: _ERROR_SCORE_TYPE
     progress_bar: bool
 
     gs_results_: Dict[str, Any]
@@ -304,11 +316,10 @@ class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT,
         pipeline: PipelineT,
         parameter_grid: ParameterGrid,
         *,
-        scoring: ScorerTypes[PipelineT, DatasetT, ScoreTypeT] = None,
+        scoring: ScorerTypes[PipelineT, DatasetT, T] = None,
         n_jobs: Optional[int] = None,
         return_optimized: Union[bool, str] = True,
         pre_dispatch: Union[int, str] = "n_jobs",
-        error_score: _ERROR_SCORE_TYPE = np.nan,
         progress_bar: bool = True,
     ) -> None:
         self.pipeline = pipeline
@@ -317,7 +328,6 @@ class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT,
         self.n_jobs = n_jobs
         self.pre_dispatch = pre_dispatch
         self.return_optimized = return_optimized
-        self.error_score = error_score
         self.progress_bar = progress_bar
 
     def optimize(self, dataset: DatasetT, **_: Any) -> Self:
@@ -357,7 +367,6 @@ class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT,
                     return_parameters=True,
                     return_data_labels=True,
                     return_times=True,
-                    error_score=self.error_score,
                 )
                 for paras in self.parameter_grid
             )
@@ -434,7 +443,7 @@ class GridSearch(BaseOptimize[PipelineT, DatasetT], Generic[PipelineT, DatasetT,
         return results
 
 
-class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[OptimizablePipelineT, DatasetT, ScoreTypeT]):
+class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[OptimizablePipelineT, DatasetT, T]):
     """Exhaustive (hyper)parameter search using a cross validation based score to optimize pipeline parameters.
 
     This class follows as much as possible the interface of :func:`~sklearn.model_selection.GridSearchCV`.
@@ -503,21 +512,25 @@ class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[Optimiz
     pre_dispatch
         The number of jobs that should be pre dispatched.
         For an explanation see the documentation of :class:`~sklearn.model_selection.GridSearchCV`
-    error_score
-        Value to assign to the score if an error occurs during scoring.
-        If set to ‘raise’, the error is raised.
-        If a numeric value is given, a Warning is raised.
     progress_bar
         True/False to enable/disable a tqdm progressbar.
     safe_optimize
         If True, we add additional checks to make sure the `self_optimize` method of the pipeline is correctly
         implemented.
         See :func:`~tpcp.make_optimize_safe` for more info.
+    optimize_with_info
+        If True, Optimize will try to call `self_optimize_with_info` by default and will fall back to `self_optimize`.
+        If you want to force the optimization to use `self_optimize`, even if an implementation of
+        `self_optimize_with_info` exists, set this parameter to False.
 
     Other Parameters
     ----------------
     dataset
         The dataset instance passed to the optimize method
+    groups
+        The groups passed to the optimize method
+    mock_labels
+        The mock labels passed to the optimize method
 
     Attributes
     ----------
@@ -586,7 +599,7 @@ class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[Optimiz
 
     pipeline: OptimizablePipelineT
     parameter_grid: ParameterGrid
-    scoring: ScorerTypes[OptimizablePipelineT, DatasetT, ScoreTypeT]
+    scoring: ScorerTypes[OptimizablePipelineT, DatasetT, T]
     return_optimized: Union[bool, str]
     cv: Optional[Union[int, BaseCrossValidator, Iterator]]
     pure_parameters: Union[bool, List[str]]
@@ -594,9 +607,12 @@ class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[Optimiz
     verbose: int
     n_jobs: Optional[int]
     pre_dispatch: Union[int, str]
-    error_score: _ERROR_SCORE_TYPE
     progress_bar: bool
     safe_optimize: bool
+    optimize_with_info: bool
+
+    groups: Optional[List[Union[str, Tuple[str, ...]]]]
+    mock_labels: Optional[List[Union[str, Tuple[str, ...]]]]
 
     cv_results_: Dict[str, Any]
     best_params_: Dict[str, Any]
@@ -610,7 +626,7 @@ class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[Optimiz
         pipeline: OptimizablePipelineT,
         parameter_grid: ParameterGrid,
         *,
-        scoring: ScorerTypes[OptimizablePipelineT, DatasetT, ScoreTypeT] = None,
+        scoring: ScorerTypes[OptimizablePipelineT, DatasetT, T] = None,
         return_optimized: Union[bool, str] = True,
         cv: Optional[Union[int, BaseCrossValidator, Iterator]] = None,
         pure_parameters: Union[bool, List[str]] = False,
@@ -618,9 +634,9 @@ class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[Optimiz
         verbose: int = 0,
         n_jobs: Optional[int] = None,
         pre_dispatch: Union[int, str] = "n_jobs",
-        error_score: _ERROR_SCORE_TYPE = np.nan,
         progress_bar: bool = True,
         safe_optimize: bool = True,
+        optimize_with_info: bool = True,
     ) -> None:
         self.pipeline = pipeline
         self.parameter_grid = parameter_grid
@@ -632,20 +648,41 @@ class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[Optimiz
         self.verbose = verbose
         self.n_jobs = n_jobs
         self.pre_dispatch = pre_dispatch
-        self.error_score = error_score
         self.progress_bar = progress_bar
         self.safe_optimize = safe_optimize
+        self.optimize_with_info = optimize_with_info
 
-    def optimize(self, dataset: DatasetT, *, groups=None, **optimize_params) -> Self:  # noqa: arguments-differ
+    def optimize(
+        self, dataset: DatasetT, *, groups=None, mock_labels=None, **optimize_params
+    ) -> Self:  # noqa: arguments-differ
+        """Run the GridSearchCV on the given dataset.
+
+        Parameters
+        ----------
+        dataset
+            The dataset to optimize on.
+        groups
+            An optional set of group labels that are passed to the cross-validation helper.
+        mock_labels
+            An optional set of mocked labels that are passed to the cross-validation helper as the `y` parameter.
+            This can be helpful in combination with the `Stratified*Fold` cross-validation helpers, that use the `y`
+            parameter to stratify the folds.
+
+        """
         self.dataset = dataset
+        self.groups = groups
+        self.mock_labels = mock_labels
+
         scoring = _validate_scorer(self.scoring, self.pipeline)
 
         cv_checked: BaseCrossValidator = check_cv(self.cv, None, classifier=True)
-        n_splits = cv_checked.get_n_splits(dataset, groups=groups)
+        n_splits = cv_checked.get_n_splits(dataset, mock_labels, groups=groups)
 
         # We need to wrap our pipeline for a consistent interface.
         # In the future we might be able to allow objects with optimizer Interface as input directly.
-        optimizer = Optimize(self.pipeline, safe_optimize=self.safe_optimize)
+        optimizer = Optimize(
+            self.pipeline, safe_optimize=self.safe_optimize, optimize_with_info=self.optimize_with_info
+        )
 
         # For each para combi, we separate the pure parameters (parameters that do not affect the optimization) and
         # the hyperparameters.
@@ -659,13 +696,15 @@ class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[Optimiz
             pure_parameters = self.pure_parameters
         else:
             raise ValueError(
-                "`self.pure_parameters` must either be a List of field names (nested are allowed) or " "True/False."
+                "`self.pure_parameters` must either be a List of field names (nested are allowed) or True/False."
             )
 
         parameters = list(self.parameter_grid)
         split_parameters = _split_hyper_and_pure_parameters(parameters, pure_parameters)
         parameter_prefix = "pipeline__"
-        combinations = list(product(enumerate(split_parameters), enumerate(cv_checked.split(dataset, groups=groups))))
+        combinations = list(
+            product(enumerate(split_parameters), enumerate(cv_checked.split(dataset, mock_labels, groups=groups)))
+        )
 
         pbar: Optional[tqdm] = None
         if self.progress_bar:
@@ -699,7 +738,6 @@ class GridSearchCV(BaseOptimize[OptimizablePipelineT, DatasetT], Generic[Optimiz
                         return_parameters=False,
                         return_data_labels=True,
                         return_times=True,
-                        error_score=self.error_score,
                         memory=tmp_cache,
                     )
                     for (cand_idx, (hyper_paras, pure_paras)), (split_idx, (train, test)) in combinations
